#include <windows.h>
#include <gl/glut.h>
#include <gl/gl.h>
#include <gl/glu.h>

#include "matrix4.h"

#include "timer.h"

#include "cell.h"

#include "stdarg.h"

#include <iostream>



CTimer timer;

Vec3 mousep;

bool updown, downdown, leftdown, rightdown, running;

Vec3 campos;

int wWidth = 600;
int wHeight = 750;

wcCA canvas(60, 10);

float simTime = 0;
int steps = 0;

float fps = 0;
float sps = 0;

float speed = -1;

int mainWindow, canvasWindow, toolWindow;

void renderCanvas();
void renderTools();

int leftButtonDown = 0;
int rightButtonDown = 0;
int middleButtonDown = 0;

int mouseX, mouseY;

int inWindow;

float waterAmount = 0.1;

int showDebug = 0;
int showHelp = 0;

float dilution = 100;

#define WIN_CANVAS 30101
#define WIN_TOOLS 30102

colCMY selectedColour(0, 1, 1);

void renderString(void* font, const char* string)
{
	char* p = (char*) string;
	while (*p != '\0') glutBitmapCharacter(font, *p++);
}

void rPrintf(float posx, float posy, const char* format, ...)
{
	va_list ap;
	char s[1024];

	va_start(ap, format);
	vsprintf((char*) s, format, ap);
	va_end(ap);

	glRasterPos2f(posx, posy);

	renderString(GLUT_BITMAP_HELVETICA_12, s);

}

void renderScene(void)
{
	
    // update camera position
	//if (updown) campos.y +=1;
	//if (downdown) campos.y -=1;
	//if (leftdown) campos.x -=1;
	//if (rightdown) campos.x +=1;

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	glTranslatef(-campos.x, -campos.y, 0);

	
	
	timer.stoptimer();
	float timeElapsed = timer.timeintv();
	timer.starttimer();

	simTime += fabs(timeElapsed);

    fps = 1/fabs(timeElapsed);


	// update the simulation
	if (simTime > 1/speed && canvas.running)
	{
		canvas.update();
		simTime = 0;
		steps++;
	}


	// draw everything
	glutSetWindow(canvasWindow);
	renderCanvas();
	
    glutSetWindow(toolWindow);
	renderTools();
	


	


}

void renderCanvas()
{
	glutSetWindow(canvasWindow);
	glClearColor(1, 1, 1, 1);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	glTranslatef(-campos.x, -campos.y, 0);

	if (leftButtonDown && inWindow == WIN_CANVAS)
	{
		for (int y = -10; y < 10; y++)
		{
			for (int x = -10; x<10; x++)
			{
				canvas.touchCell(mouseX+x, mouseY+y, selectedColour/dilution, waterAmount);
			}			
		}
	}

	if (middleButtonDown && inWindow == WIN_CANVAS)
	{
		for (int y = -10; y < 10; y++)
		{
			for (int x = -10; x<10; x++)
			{
				canvas.touchCell(mouseX+x, mouseY+y, colCMY(-0.05, -0.05, -0.05), -waterAmount);
			}			
		}
	}

	if (rightButtonDown && inWindow == WIN_CANVAS)
	{
		for (int y = -10; y < 10; y++)
		{
			for (int x = -10; x<10; x++)
			{
				canvas.touchCell(mouseX+x, mouseY+y, colCMY(0, 0, 0), waterAmount);
			}			
		}
	}

	canvas.draw();

	// draw any text we might want
	glColor3f(0, 0, 0);
	if (showDebug)
	{
		rPrintf(10, 600-20, "Debug: step %d", steps);
		rPrintf(10, 600-50, "cEvaporation     %.4f", canvas.cEvaporation);
		rPrintf(10, 600-65, "cSpreading       %.4f", canvas.cSpreading);
		rPrintf(10, 600-80, "cSurfaceTension  %.4f", canvas.cSurfaceTension);
		rPrintf(10, 600-95, "cGravity         %.4f", canvas.cGravity);
		rPrintf(10, 600-110,"cAbsorptionRate  %.4f", canvas.cAbsorptionRate);
		rPrintf(10, 600-125,"Water amount     %.4f", waterAmount);
		rPrintf(10, 600-140,"Pigment dilution %.4f", dilution);
		
		if (inWindow == WIN_CANVAS)
		{
			wcCell* curCell = &canvas.mCellArray[(mouseY/canvas.dCellSize) * canvas.caSize + (mouseX/canvas.dCellSize)];
			rPrintf(10, 600-100-140, "Cell:");
			rPrintf(40, 600-100-155, "fluidContent    %.4f", curCell->fluidContent);
			rPrintf(40, 600-100-170, "fluidDX         %.4f", curCell->fluidDX);
			rPrintf(40, 600-100-185, "fluidDY         %.4f", curCell->fluidDY);
			rPrintf(40, 600-100-200, "C               %.4f", curCell->pigmentContent.c);
			rPrintf(40, 600-100-215, "M               %.4f", curCell->pigmentContent.m);
			rPrintf(40, 600-100-230, "Y               %.4f", curCell->pigmentContent.y);
			rPrintf(40, 600-100-245, "Ci              %.4f", curCell->pigmentInfused.c);
			rPrintf(40, 600-100-260, "Mi              %.4f", curCell->pigmentInfused.m);
			rPrintf(40, 600-100-275, "Yi              %.4f", curCell->pigmentInfused.y);
			rPrintf(40, 600-100-290, "fluidInfused    %.4f", curCell->fluidInfused);
			rPrintf(40, 600-100-305, "Absorbency      %.4f", canvas.absorbencyField[(mouseY/canvas.dCellSize) * canvas.caSize + (mouseX/canvas.dCellSize)]);
		}

		rPrintf(10, 10, "FPS %.1f", fps);


	}
	if (showHelp)
	{
		rPrintf(300, 600-20, "Help:");
		rPrintf(300, 600-50, "Q\\A - Surface tension");
		rPrintf(300, 600-65, "W\\S - Spreading");
		rPrintf(300, 600-80, "E\\D - Absorption rate");
		rPrintf(300, 600-95, "R\\F - Evaporation");
		rPrintf(300, 600-110, "T\\G - Water quantity");
		rPrintf(300, 600-125, "Y\\H - Pigment evaporation");
		rPrintf(300, 300, "F2  - Show debug info");
		rPrintf(300, 300-15, "F5  - Toggle pigment/solution/fluid display");
		rPrintf(300, 300-30, "F9  - Clear canvas");
		rPrintf(300, 300-45, "F10 - Run simulation");
		rPrintf(300, 300-60, "F11 - Step through simulation");
	}
	

	

	glutSwapBuffers();
}

void normalKeys(unsigned char key, int x, int y)
{
	switch (key)
	{
	case 'q':
		canvas.cSurfaceTension += 0.001;
		break;
	case 'a':
		canvas.cSurfaceTension -= 0.001;
		break;
	case 'w':
		canvas.cSpreading += 0.01;
		break;
	case 's':
		canvas.cSpreading -= 0.01;
		break;
	case 'e':
		canvas.cAbsorptionRate += 0.01;
		break;
	case 'd':
		canvas.cAbsorptionRate -= 0.01;
		break;
	case 'r':
		canvas.cEvaporation += 0.01;
		break;
	case 'f':
		canvas.cEvaporation -= 0.01;
		break;
	case 't':
		waterAmount += 0.1;
		break;
	case 'g':
		waterAmount -= 0.1;
		break;
	case 'y':
		dilution += 5;
		break;
	case 'h':
		dilution -= 5;
		break;
	}
}

#define BARWIDTH 30
#define BARLENGTH 150
#define BORDER 12
#define BUDGE 10

void renderTools()
{
	glutSetWindow(toolWindow);
	glClearColor(1, 1, 1, 1);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	

	// draw colour selectors
	

	glBegin(GL_QUADS);
	// red bar
		glColor3f(1,1-selectedColour.m,1-selectedColour.y);
		glVertex2i(BORDER, BORDER + BUDGE);
		glVertex2i(BORDER, BORDER+BARWIDTH + BUDGE);
		glColor3f(0,1-selectedColour.m,1-selectedColour.y);
		glVertex2i(BORDER+BARLENGTH, BORDER+BARWIDTH + BUDGE);
		glVertex2i(BORDER+BARLENGTH, BORDER + BUDGE);
	// green bar
		glColor3f(1-selectedColour.c,1,1-selectedColour.y);
		glVertex2i(BORDER, BORDER*2+BARWIDTH + BUDGE);
		glVertex2i(BORDER, BORDER*2+BARWIDTH*2 + BUDGE);
		glColor3f(1-selectedColour.c,0,1-selectedColour.y);
		glVertex2i(BORDER+BARLENGTH, BORDER*2+BARWIDTH*2 + BUDGE);
		glVertex2i(BORDER+BARLENGTH, BORDER*2+BARWIDTH + BUDGE);
	// blue bar	
		glColor3f(1-selectedColour.c,1-selectedColour.m,1);
		glVertex2i(BORDER, BORDER*3 + BARWIDTH*2 + BUDGE);
		glVertex2i(BORDER, BORDER*3 + BARWIDTH*3 + BUDGE);
		glColor3f(1-selectedColour.c,1-selectedColour.m,0);
		glVertex2i(BORDER+BARLENGTH, BORDER*3 + BARWIDTH*3 + BUDGE);
		glVertex2i(BORDER+BARLENGTH, BORDER*3 + BARWIDTH*2 + BUDGE);

		// show selected colour
		glColor3f(1-selectedColour.c,1-selectedColour.m,1-selectedColour.y);
		//glColor3f(0, 0, 0);
		glVertex2i(200, 30);
		glVertex2i(300, 30);
		glVertex2i(300, 130);
		glVertex2i(200, 130);
		/*
		glColor3f(0, 1, 0);
		glVertex2i(mouseX, mouseY);
		glVertex2i(mouseX+2, mouseY);
		glVertex2i(mouseX+2, mouseY+2);
		glVertex2i(mouseX, mouseY+2);*/

	glEnd();

	// check to see if we're clicking on a colour
	if (leftButtonDown && inWindow == WIN_TOOLS)
	{
		//if within bar range
		if (mouseX >= BORDER && mouseX <= BORDER+BARLENGTH)
		{
			//if in the red bar
			if (mouseY >= BORDER+BUDGE && mouseY <= BORDER+BARWIDTH+BUDGE)
			{
				selectedColour.c = ((float)mouseX - (float)BORDER) / (float)BARLENGTH;
			}

			//if in the green bar
			if (mouseY >= BORDER*2+BARWIDTH + BUDGE && mouseY <= BORDER*2+BARWIDTH*2 + BUDGE)
			{
				selectedColour.m = ((float)mouseX - (float)BORDER) / (float)BARLENGTH;
			}

			//if in the blue bar
			if (mouseY >= BORDER*3 + BARWIDTH*2 + BUDGE && mouseY <= BORDER*3 + BARWIDTH*3 + BUDGE)
			{
				selectedColour.y = ((float)mouseX - (float)BORDER) / (float)BARLENGTH;
			}
		}

		//std::cout << selectedColour.c << ", " << selectedColour.m << ", "<< selectedColour.y << std::endl;
	}



	glutSwapBuffers();

}


void initGL(void)
{
	glClearColor(1, 1, 1, 1);
	glColor3f(0, 0, 0);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	//gluPerspective(45, 1, 0.1, 100);
	gluOrtho2D(0, wWidth, 0, wHeight);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

}

void initCanvas(void)
{
	glClearColor(1, 1, 1, 1);
	glColor3f(0, 0, 0);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	//gluPerspective(45, 1, 0.1, 100);
	gluOrtho2D(0, 600, 0, 600);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

}

void initTools(void)
{
	glClearColor(1, 1, 1, 1);
	glColor3f(0, 0, 0);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	//gluPerspective(45, 1, 0.1, 100);
	gluOrtho2D(0, 600, 0, 150);
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

}

void mouseHandler(int button, int state, int x, int y)
{
	if (state == GLUT_UP)
	{
		switch (button)
		{
		case GLUT_LEFT_BUTTON:
			leftButtonDown = 0;
			break;
		case GLUT_MIDDLE_BUTTON:
			middleButtonDown = 0;
			break;
		case GLUT_RIGHT_BUTTON:
			rightButtonDown = 0;
			break;
		}
	}
	if (state == GLUT_DOWN)
	{
		switch (button)
		{
		case GLUT_LEFT_BUTTON:
			leftButtonDown = 1;
			break;
		case GLUT_MIDDLE_BUTTON:
			middleButtonDown = 1;
			break;
		case GLUT_RIGHT_BUTTON:
			rightButtonDown = 1;
			break;
		}
	}

	mouseX = x; mouseY = y;
}

void pMouse(int x, int y)
{
	mouseX = x; mouseY = y;
}

void mouseHandlerC(int button, int state, int x, int y)
{
	if (state == GLUT_UP)
	{
		switch (button)
		{
		case GLUT_LEFT_BUTTON:
			leftButtonDown = 0;
			break;
		case GLUT_MIDDLE_BUTTON:
			middleButtonDown = 0;
			break;
		case GLUT_RIGHT_BUTTON:
			rightButtonDown = 0;
			break;
		}
	}
	if (state == GLUT_DOWN)
	{
		switch (button)
		{
		case GLUT_LEFT_BUTTON:
			leftButtonDown = 1;
			break;
		case GLUT_MIDDLE_BUTTON:
			middleButtonDown = 1;
			break;
		case GLUT_RIGHT_BUTTON:
			rightButtonDown = 1;
			break;
		}
	}

	mouseX = x; mouseY = 600-y;

	inWindow = WIN_CANVAS;
	
}

void mouseMotionC(int x, int y)
{
	mouseX = x, mouseY = 600-y;
	inWindow = WIN_CANVAS;
}

void pMouseC(int x, int y)
{
	mouseX = x; mouseY = 600-y;

	inWindow = WIN_CANVAS;
	
}

void mouseHandlerT(int button, int state, int x, int y)
{
	if (state == GLUT_UP)
	{
		switch (button)
		{
		case GLUT_LEFT_BUTTON:
			leftButtonDown = 0;
			break;
		case GLUT_MIDDLE_BUTTON:
			middleButtonDown = 0;
			break;
		case GLUT_RIGHT_BUTTON:
			rightButtonDown = 0;
			break;
		}
	}
	if (state == GLUT_DOWN)
	{
		switch (button)
		{
		case GLUT_LEFT_BUTTON:
			leftButtonDown = 1;
			break;
		case GLUT_MIDDLE_BUTTON:
			middleButtonDown = 1;
			break;
		case GLUT_RIGHT_BUTTON:
			rightButtonDown = 1;
			break;
		}
	}

	mouseX = x; mouseY = 150-y;

	inWindow = WIN_TOOLS;
	
}

void pMouseT(int x, int y)
{
	mouseX = x; mouseY = 150-y;

	inWindow = WIN_TOOLS;
	
}

void keyUp(int key, int x, int y) {
	switch (key)
	{
	case GLUT_KEY_LEFT:
		leftdown = false;
		break;
	case GLUT_KEY_RIGHT:
		rightdown = false;
		break;
	case GLUT_KEY_DOWN:
		downdown = false;
		break;
	case GLUT_KEY_UP:
		updown = false;
		break;
	case GLUT_KEY_F10:
		canvas.running = !canvas.running;
		break;
	case GLUT_KEY_F5:
		canvas.displayMode += 1;
		if (canvas.displayMode > 2) canvas.displayMode = 0;
		break;
	case GLUT_KEY_F9:
		canvas.clear();
		steps = 0;
		break;
	case GLUT_KEY_F11:
		if (!canvas.running) canvas.update();
		steps++;
		break;
	case GLUT_KEY_F1:
		showHelp = !showHelp;
		break;
	case GLUT_KEY_F2:
		showDebug = !showDebug;
		break;

	}
}

void keyDown(int key, int x, int y) {
	switch (key)
	{
	case GLUT_KEY_LEFT:
		leftdown = true;
		break;
	case GLUT_KEY_RIGHT:
		rightdown = true;
		break;
	case GLUT_KEY_DOWN:
		downdown = true;
		break;
	case GLUT_KEY_UP:
		updown = true;
		break;

	}
}

void menuFunc(int value)
{
	if (value == -1) canvas.clear();
	
	if (value == 10030) speed = 0;
	else speed = 1/(value-10000);
}




int main (int argc, char** argv)
{
	glutInit(&argc, argv);
	glutInitWindowPosition(100, 100);
	glutInitWindowSize(wWidth, wHeight);
	glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);
	
	// main window
	mainWindow = glutCreateWindow("Watercolour 0.1");
		glutDisplayFunc(renderScene);
		glutMouseFunc(mouseHandler);
		glutPassiveMotionFunc(pMouse);
		glutIdleFunc(renderScene);
		
		glutSpecialFunc(keyDown);
		glutSpecialUpFunc(keyUp);
		
		glutKeyboardUpFunc(normalKeys);
		

		initGL();


	canvasWindow = glutCreateSubWindow(mainWindow, 0, 0, 600, 600);
		glutDisplayFunc(renderCanvas);
		initCanvas();
		glutMouseFunc(mouseHandlerC);
		glutMotionFunc(mouseMotionC);
		glutPassiveMotionFunc(pMouseC);
		glutSpecialFunc(keyDown);
		glutSpecialUpFunc(keyUp);
		glutKeyboardUpFunc(normalKeys);

	toolWindow = glutCreateSubWindow(mainWindow, 0, 600, 600, 150);
		glutDisplayFunc(renderTools);
		initTools();
		glutMouseFunc(mouseHandlerT);
		glutPassiveMotionFunc(pMouseT);
		glutSpecialFunc(keyDown);
		glutSpecialUpFunc(keyUp);
		glutKeyboardUpFunc(normalKeys);

	// scene init
	
	updown = downdown = leftdown = rightdown = false;
	
	canvas.init();

	timer.starttimer();


	glutMainLoop();
}