#pragma once

#include <math.h>

#include <iostream>

inline
float cub(float n){
	return n*n*n;
}

inline
float sqr(float n){
	return n*n;
}

class Vec3
{
public:
	Vec3(float _x = 0, float _y = 0, float _z = 0){ x = _x; y = _y; z = _z;}
	Vec3(Vec3& e){x = e.x; y = e.y; z = e.z;}
	~Vec3(){};

	float x, y, z;

	inline void set(float _x, float _y, float _z) { x = _x; y = _y; z = _z;}



	inline Vec3 operator -(Vec3& s){
		Vec3 r;
		r.x = x - s.x;
		r.y = y - s.y;
		r.z = z - s.z;
		return r;
	}

	inline Vec3 operator *(float s){
		Vec3 r;
		r.x = x*s;
		r.y = y*s;
		r.z = z*s;
		return r;
	}

	inline Vec3 operator +(Vec3& a){
		Vec3 r;
		r.x = x + a.x;
		r.y = y + a.y;
		r.z = z + a.z;
		return r;
	}

	inline Vec3 operator / (float f) {
		Vec3 r;
		r.x = x / f;
		r.y = y / f;
		r.z = z / f;
		return r;
	}

	inline Vec3 operator ^(Vec3& c){
		Vec3 result;
		result.x = y * c.z - c.y * z;
		result.y = z * c.x - c.z * x;
		result.z = x * c.y - c.x * y;
		return result;
	}

	inline void operator =(Vec3& e){
		x = e.x;
		y = e.y;
		z = e.z;
	}

	inline float operator *(Vec3& d){
		return x*d.x + y*d.y + z*d.z;
	}

	inline bool operator == (Vec3& e) {
		return (fabs(x-e.x) < 0.05) && (fabs(y-e.y) < 0.05) && (fabs(z-e.z) < 0.05) ? true : false;
	}

	inline bool operator != (Vec3& e) {
		return !(*this == e);
	}


	inline void operator += (Vec3 v) {
		*this = *this + v;
	}

	inline void operator -= (Vec3& v) {
		*this = *this - v;
	}

	inline void operator *= (float f) {
		*this = *this * f;
	}

	inline void operator /= (float f) {
		*this = *this / f;
	}

	inline float length()
	{
		return sqrt(sqr(x) + sqr(y) + sqr(z));
	}

	inline Vec3 normal()
	{
		return *this / length();
	}

	inline void normalize()
	{
		*this = *this / length();
	}

	inline void truncate(float t) {
		if (fabs(length()) > t)
			*this = normal() * t;
	}

	
};

inline std::ostream& operator << (std::ostream& os, Vec3 v) {
	os << v.x << " " << v.y << " " << v.z;
	return os;
}

inline Vec3 operator - (Vec3 v) {
	Vec3 r;
	r.x = -v.x;
	r.y = -v.y;
	r.z = -v.z;
	return r;
}

