#pragma once
#include <windows.h>
#include <gl/gl.h>
#include <math.h>
#include <stdio.h>
#include "vector3.h"



#define PI 3.1415926535897932384626433832795
#define PIOVERONEEIGHTY 0.017453292519943295769236907684883
#define ONEEIGHTYOVERPI 57.295779513082320876798154814114

#define DEG2RAD(x) (x * PIOVERONEEIGHTY)
#define RAD2DEG(x) (x * ONEEIGHTYOVERPI)

#define X 0
#define Y 1
#define Z 2
#define W 3

inline int Random(int lb, int hb) {
	return rand()%(hb-lb) + lb;
}

class Mat4
{
public:
	float _00, _10, _20, _30;
	float _01, _11, _21, _31;
	float _02, _12, _22, _32;
	float _03, _13, _23, _33;

	void		loadGLMatrix(GLenum matrixname);
	void		loadIdentity(void);

	inline Mat4	inverse(void);
	inline Mat4	cofactor(void);
	inline Mat4	transpose(void);
	inline Mat4	rotateDeg(float angle, int axis);
	inline Mat4	rotateTrig(float sinx, float cosx, int axis);
	inline Mat4	translate(float x, float y, float z);
	inline Mat4 scale(float x, float y, float z);
	inline Mat4	multScalar(float s);
	inline Mat4	multMatrix(Mat4& m);
	inline Mat4 operator * (Mat4& m) {
		return multMatrix(m);
	}

	inline float& operator () (int col, int row) {
		if ((col < 0 || col > 3) || (row < 0 || row > 3))
			throw;
		else
		{
			switch (col)
			{
			case 0:
				switch (row)
				{
				case 0:
					return _00;
					break;
				case 1:
					return _01;
					break;
				case 2:
					return _02;
					break;
				case 3:
					return _03;
					break;
				}
			case 1:
				switch (row)
				{
				case 0:
					return _10;
					break;
				case 1:
					return _11;
					break;
				case 2:
					return _12;
					break;
				case 3:
					return _13;
					break;
				}
			case 2:
				switch (row)
				{
				case 0:
					return _20;
					break;
				case 1:
					return _21;
					break;
				case 2:
					return _22;
					break;
				case 3:
					return _23;
					break;
				}
			case 3:
				switch (row)
				{
				case 0:
					return _30;
					break;
				case 1:
					return _31;
					break;
				case 2:
					return _32;
					break;
				case 3:
					return _33;
					break;
				}
			}
		}
	}

	inline Vec3 operator * (Vec3& v) {
		Vec3 r;
		r.x = _00*v.x + _10*v.y + _20*v.z;
		r.y = _01*v.x + _11*v.y + _21*v.z;
		r.z = _02*v.x + _12*v.y + _22*v.z;

		return r;
	}

	
	Vec3 multAsPoint(Vec3& p) {
		Vec3 pt;
		pt.x =		_00*p.x + _10*p.y + _20*p.z + _30;
		pt.y =		_01*p.x + _11*p.y + _21*p.z + _31;
		pt.z =		_02*p.x + _12*p.y + _22*p.z + _32;
		float w =	_03*p.x + _13*p.y + _23*p.z + _33;

		return pt / w;
	}

	void print();

	int singular;		
	float		det(void);
	Mat4();
};


class Mat3
{
public:
	float _00, _10, _20;
	float _01, _11, _21;
	float _02, _12, _22;

	inline Mat3 inverse();
	inline Mat3 cofactor();
	inline Mat3 transpose();
	inline Mat3 multScalar(float s);

	int singular;

	void print();

	float det(void);
	Mat3();
};





Mat3::Mat3()
{
	_00 = 1; _10 = 0; _20 = 0;
	_01 = 0; _11 = 1; _21 = 0;
	_02 = 0; _12 = 0; _22 = 1;

	singular = 0;
}

float Mat3::det()
{
	return _00*(_11*_22 - _12*_21) - _10*(_01*_22 - _02*_21) + _20*(_01*_12 - _02*_11);
}


Mat3 Mat3::cofactor()
{
	Mat3 r;

	r._00 = (_11*_22 - _12*_21); r._10 = -(_01*_22 - _02*_21); r._20 = (_01*_12 - _02*_11);
	r._01 = -(_10*_22 - _12*_20); r._11 = (_00*_22 - _02*_20); r._21 = -(_00*_12 - _02*_10);
	r._02 = (_10*_21 - _11*_20); r._12 = -(_00*_21 - _01*_20); r._22 = (_00*_11 - _01*_10);

	return r;
}

Mat3 Mat3::transpose()
{
	Mat3 r;

	r._00 = _00; r._10 = _01; r._20 = _02;
	r._01 = _10; r._11 = _11; r._21 = _12;
	r._02 = _20; r._12 = _21; r._22 = _22;

	return r;
}


Mat3 Mat3::multScalar(float s)
{
	Mat3 r;

	r._00 = _00*s; r._10 = _10*s; r._20 = _20*s;
	r._01 = _01*s; r._11 = _11*s; r._21 = _21*s;
	r._02 = _02*s; r._12 = _12*s; r._22 = _22*s;

	return r;
}

Mat3 Mat3::inverse()
{
	Mat3 r;

	float det = this->det();

	r = this->cofactor().transpose();

	return r.multScalar(1/det);
}

void Mat3::print()
{
	printf("%2.2f %2.2f %2.2f\n", _00, _10, _20);
	printf("%2.2f %2.2f %2.2f\n", _01, _11, _21);
	printf("%2.2f %2.2f %2.2f\n", _02, _12, _22);

	printf("\n");
}



Mat4::Mat4()
{
	//just load the identity matrix
	_00 = 1; _10 = 0; _20 = 0; _30 = 0;
	_01 = 0; _11 = 1; _21 = 0; _31 = 0;
	_02 = 0; _12 = 0; _22 = 1; _32 = 0;
	_03 = 0; _13 = 0; _23 = 0; _33 = 1;

	singular = 0;
}

void Mat4::loadIdentity()
{
	//just load the identity matrix
	_00 = 1; _10 = 0; _20 = 0; _30 = 0;
	_01 = 0; _11 = 1; _21 = 0; _31 = 0;
	_02 = 0; _12 = 0; _22 = 1; _32 = 0;
	_03 = 0; _13 = 0; _23 = 0; _33 = 1;
}

void Mat4::loadGLMatrix(GLenum matrixname)
{
	double matrix[16];
	glGetDoublev(matrixname, matrix);

	_00 = matrix[ 0]; _10 = matrix[ 4]; _20 = matrix[ 8]; _30 = matrix[12];
	_01 = matrix[ 1]; _11 = matrix[ 5]; _21 = matrix[ 9]; _31 = matrix[13];
	_02 = matrix[ 2]; _12 = matrix[ 6]; _22 = matrix[10]; _32 = matrix[14];
	_03 = matrix[ 3]; _13 = matrix[ 7]; _23 = matrix[11]; _33 = matrix[15];
}


float Mat4::det()
{
	//get the submatrices
	Mat3 a, e, i, m;

	a._00 = _11; a._10 = _21; a._20 = _31;
	a._01 = _12; a._11 = _22; a._21 = _32;
	a._02 = _13; a._12 = _23; a._22 = _33;

	e._00 = _01; e._10 = _21; e._20 = _31;
	e._01 = _02; e._11 = _22; e._21 = _32;
	e._02 = _03; e._12 = _23; e._22 = _33;

	i._00 = _01; i._10 = _11; i._20 = _31;
	i._01 = _02; i._11 = _12; i._21 = _32;
	i._02 = _03; i._12 = _13; i._22 = _33;

	m._00 = _01; m._10 = _11; m._20 = _21;
	m._01 = _02; m._11 = _12; m._21 = _22;
	m._02 = _03; m._12 = _13; m._22 = _23;

	return _00*a.det() - _10*e.det() + _20*i.det() - _30*m.det();
	
}

Mat4 Mat4::cofactor()
{
	Mat3 a, b, c, d, e, f, g, h, i, j, k, l, m, n, o, p;
	Mat4 r;

	a._00 = _11; a._10 = _21; a._20 = _31;		e._00 = _01; e._10 = _21; e._20 = _31;		i._00 = _01; i._10 = _11; i._20 = _31;		m._00 = _01; m._10 = _11; m._20 = _21;
	a._01 = _12; a._11 = _22; a._21 = _32;		e._01 = _02; e._11 = _22; e._21 = _32;		i._01 = _02; i._11 = _12; i._21 = _32;		m._01 = _02; m._11 = _12; m._21 = _22;
	a._02 = _13; a._12 = _23; a._22 = _33;		e._02 = _03; e._12 = _23; e._22 = _33;		i._02 = _03; i._12 = _13; i._22 = _33;		m._02 = _03; m._12 = _13; m._22 = _23;
																								
	b._00 = _10; b._10 = _20; b._20 = _30;		f._00 = _00; f._10 = _20; f._20 = _30;		j._00 = _00; j._10 = _10; j._20 = _30;		n._00 = _00; n._10 = _10; n._20 = _20;
	b._01 = _12; b._11 = _22; b._21 = _32;		f._01 = _02; f._11 = _22; f._21 = _32;		j._01 = _02; j._11 = _12; j._21 = _32;		n._01 = _02; n._11 = _12; n._21 = _22;
	b._02 = _13; b._12 = _23; b._22 = _33;		f._02 = _03; f._12 = _23; f._22 = _33;		j._02 = _03; j._12 = _13; j._22 = _33;		n._02 = _03; n._12 = _13; n._22 = _23;
																								
	c._00 = _10; c._10 = _20; c._20 = _30;		g._00 = _00; g._10 = _20; g._20 = _30;		k._00 = _00; k._10 = _10; k._20 = _30;		o._00 = _00; o._10 = _10; o._20 = _20;
	c._01 = _11; c._11 = _21; c._21 = _31;		g._01 = _01; g._11 = _21; g._21 = _31;		k._01 = _01; k._11 = _11; k._21 = _31;		o._01 = _01; o._11 = _11; o._21 = _21;
	c._02 = _13; c._12 = _23; c._22 = _33;		g._02 = _03; g._12 = _23; g._22 = _33;		k._02 = _03; k._12 = _13; k._22 = _33;		o._02 = _03; o._12 = _13; o._22 = _23;
																							
	d._00 = _10; d._10 = _20; d._20 = _30;		h._00 = _00; h._10 = _20; h._20 = _30;		l._00 = _00; l._10 = _10; l._20 = _30;		p._00 = _00; p._10 = _10; p._20 = _20;
	d._01 = _11; d._11 = _21; d._21 = _31;		h._01 = _01; h._11 = _21; h._21 = _31;		l._01 = _01; l._11 = _11; l._21 = _31;		p._01 = _01; p._11 = _11; p._21 = _21;
	d._02 = _12; d._12 = _22; d._22 = _32;		h._02 = _02; h._12 = _22; h._22 = _32;		l._02 = _02; l._12 = _12; l._22 = _32;		p._02 = _02; p._12 = _12; p._22 = _22;


	
	r._00 =  a.det();   r._10 = -e.det();   r._20 =  i.det();   r._30 = -m.det();
	r._01 = -b.det();   r._11 =  f.det();   r._21 = -j.det();   r._31 =  n.det();
	r._02 =  c.det();   r._12 = -g.det();   r._22 =  k.det();   r._32 = -o.det();
	r._03 = -d.det();   r._13 =  h.det();   r._23 = -l.det();   r._33 =  p.det();

	return r;
}

Mat4 Mat4::transpose(void)
{
	Mat4 r;

	r._00 = _00;  r._10 = _01;  r._20 = _02;  r._30 = _03;
	r._01 = _10;  r._11 = _11;  r._21 = _12;  r._31 = _13;
	r._02 = _20;  r._12 = _21;  r._22 = _22;  r._32 = _23;
	r._03 = _30;  r._13 = _31;  r._23 = _32;  r._33 = _33;

	return r;

}

Mat4 Mat4::multScalar(float s)
{
	Mat4 r;

	r._00 = _00*s;  r._10 = _10*s;  r._20 = _20*s; r._30 = _30*s;
	r._01 = _01*s;  r._11 = _11*s;  r._21 = _21*s; r._31 = _31*s;
	r._02 = _02*s;  r._12 = _12*s;  r._22 = _22*s; r._32 = _32*s;
	r._03 = _03*s;  r._13 = _13*s;  r._23 = _23*s; r._33 = _33*s;

	return r;

}


Mat4 Mat4::inverse()
{
	float det = this->det();
	Mat4 r;
	
	if (det == 0)
	{
		r.singular = 1;
		return r;
	}

	r = this->cofactor().transpose();

	return r.multScalar(1/det);
	
}

Mat4 Mat4::rotateDeg(float angle, int axis)
{
	Mat4 r;

	switch (axis)
	{
	case Z:
		r._00 = cos(DEG2RAD(angle)); r._10 = -sin(DEG2RAD(angle));
		r._01 = sin(DEG2RAD(angle)); r._11 = cos(DEG2RAD(angle));
		break;

	case Y:
		r._00 = cos(DEG2RAD(angle)); r._20 = sin(DEG2RAD(angle));
		r._02 = -sin(DEG2RAD(angle)); r._22 = cos(DEG2RAD(angle));
		break;

	case X:
		r._11 = cos(DEG2RAD(angle)); r._21 = -sin(DEG2RAD(angle));
		r._12 = sin(DEG2RAD(angle)); r._22 = cos(DEG2RAD(angle));
		break;
	}


	return multMatrix(r);
}


Mat4 Mat4::rotateTrig(float sinx, float cosx, int axis)
{
	Mat4 r;

	switch (axis)
	{
	case Z:
		r._00 = cosx; r._10 = -sinx;
		r._01 = sinx; r._11 = cosx;
		break;

	case Y:
		r._00 = cosx; r._20 = sinx;
		r._02 = -sinx; r._22 = cosx;
		break;

	case X:
		r._11 = cosx; r._21 = -sinx;
		r._12 = sinx; r._22 = cosx;
		break;
	}


	return multMatrix(r);
}



Mat4 Mat4::multMatrix(Mat4& m)
{
	Mat4 r;

	r._00 = m._00*_00 + m._10*_01 + m._20*_02 + m._30*_03;		r._10 = m._00*_10 + m._10*_11 + m._20*_12 + m._30*_13;		r._20 = m._00*_20 + m._10*_21 + m._20*_22 + m._30*_23;	  r._30 = m._00*_30 + m._10*_31 + m._20*_32 + m._30*_33;
	r._01 = m._01*_00 + m._11*_01 + m._21*_02 + m._31*_03;		r._11 = m._01*_10 + m._11*_11 + m._21*_12 + m._31*_13;		r._21 = m._01*_20 + m._11*_21 + m._21*_22 + m._31*_23;	  r._31 = m._01*_30 + m._11*_31 + m._21*_32 + m._31*_33;
	r._02 = m._02*_00 + m._12*_01 + m._22*_02 + m._32*_03;		r._12 = m._02*_10 + m._12*_11 + m._22*_12 + m._32*_13;		r._22 = m._02*_20 + m._12*_21 + m._22*_22 + m._32*_23;	  r._32 = m._02*_30 + m._12*_31 + m._22*_32 + m._32*_33;
	r._03 = m._03*_00 + m._13*_01 + m._23*_02 + m._33*_03;		r._13 = m._03*_10 + m._13*_11 + m._23*_12 + m._33*_13;		r._23 = m._03*_20 + m._13*_21 + m._23*_22 + m._33*_23;	  r._33 = m._03*_30 + m._13*_31 + m._23*_32 + m._33*_33;

	return r;

}


Mat4 Mat4::translate(float x, float y, float z)
{
	Mat4 r;

	r._30 = x;
	r._31 = y;
	r._32 = z;

	return multMatrix(r);
}


void Mat4::print()
{
	printf("%2.2f %2.2f %2.2f %2.2f\n", _00, _10, _20, _30);
	printf("%2.2f %2.2f %2.2f %2.2f\n", _01, _11, _21, _31);
	printf("%2.2f %2.2f %2.2f %2.2f\n", _02, _12, _22, _32);
	printf("%2.2f %2.2f %2.2f %2.2f\n", _03, _13, _23, _33);
	printf("\n");
}
