#pragma once

#include <vector>

class Cell
{
public:
	Cell(){
		state = 0;
	}

	int state;

	
};



class CA
{
public:
	CA(int _caSize = 100, int _dCellSize = 5) {
		
		mCellArray.resize(_caSize*_caSize);
		running = 0;
		caSize = _caSize;
		dCellSize = _dCellSize;
	}

	void testPattern()
	{
		for (int i = 0; i < caSize*caSize; i++)
		{
			if (i % 2 == 0)
			{
				mCellArray[i].state = 1;
			}
		}
	}

	void clear()
	{
		running = 0;
		mCellArray.clear();
		mCellArray.resize(caSize*caSize);
	}

	void update();

	void draw();

	void touchCell(int x, int y);

	std::vector<Cell> mCellArray;
	int running;
	int caSize;
	int dCellSize;

private:
	int neighbours(int x, int y);
};

void CA::draw()
{

	// draw the cell array (only draw alive cells)
	glColor3f(0, 0, 0.5);
	glBegin(GL_QUADS);
	for (int y = 0; y < caSize; y++) // outside loop is rows (y)
	{
		for (int x = 0; x < caSize; x++) // inside loop is columns (x)
		{
			if (mCellArray[y*caSize + x].state)
			{
				glVertex2i(x*dCellSize, y*dCellSize);
				glVertex2i(x*dCellSize + dCellSize, y*dCellSize);
				glVertex2i(x*dCellSize + dCellSize, y*dCellSize + dCellSize);
				glVertex2i(x*dCellSize, y*dCellSize + dCellSize);
			}
		}
	}
	glEnd();

	//draw a grid
	if (running) glColor3f(0.8, 0.8, 1.0);
	else glColor3f(0.8, 0.8, 0.8);
	
	glBegin(GL_LINES);
	for (int i = 0; i <= caSize; i++)
	{
		//vertical
		glVertex2i(i*dCellSize, 0);
		glVertex2i(i*dCellSize, caSize*dCellSize);

		//horizontal
		glVertex2i(0, i*dCellSize);
		glVertex2i(caSize*dCellSize, i*dCellSize);
	}
	glEnd();
}

void CA::touchCell(int _x, int _y)
{
	int cX = _x / dCellSize;
	int cY = _y / dCellSize;

	if (cX < caSize && cY < caSize)
	{
		mCellArray[cY*caSize + cX].state = !mCellArray[cY*caSize + cX].state;
	}
}

void CA::update()
{
	std::vector<Cell> tempArray;
	tempArray.resize(caSize*caSize);
	//std::cout << "Update" << std::endl;
	int nb;
	for (int y = 0; y < caSize; y++)
	{
		for (int x = 0; x < caSize; x++)
		{
			nb = neighbours(x, y);
			//std::cout << nb << std::endl;
			if (mCellArray[y*caSize + x].state) // if cell is alive
			{
				if (nb == 2 || nb == 3)
					tempArray[y*caSize + x].state = 1;		// if cell has 2 or 3 neighbours, survives
				else
					tempArray[y*caSize + x].state = 0;
			}
			else	// if cell is dead
			{
				if (nb == 3) tempArray[y*caSize + x].state = 1;; // if cell has 3 neighbours, born
			}
		}
	}

	mCellArray = tempArray;
}

int CA::neighbours(int x, int y)
{
	// check to see how many living neighbours cell at (x, y) has
	int num = 0;

	int xIdxPos, xIdxNeg, yIdxPos, yIdxNeg;
	
	// wrap the simulation round the grid
	xIdxPos = x + 1;
	if (xIdxPos >= caSize) xIdxPos -= caSize;
	yIdxPos = y + 1;
	if (yIdxPos >= caSize) yIdxPos -= caSize;
	xIdxNeg = x - 1;
	if (xIdxNeg < 0) xIdxNeg += caSize;
	yIdxNeg = y - 1;
	if (yIdxNeg < 0) yIdxNeg += caSize;

	num += mCellArray[yIdxPos*caSize + xIdxNeg].state; //top-left
	num += mCellArray[yIdxPos*caSize + x].state; //top-middle
	num += mCellArray[yIdxPos*caSize + xIdxPos].state; // top-right
	num += mCellArray[y*caSize + xIdxPos].state; // middle-right
	num += mCellArray[yIdxNeg*caSize + xIdxPos].state; // bottom-right
	num += mCellArray[yIdxNeg*caSize + x].state; // bottom-middle
	num += mCellArray[yIdxNeg*caSize + xIdxNeg].state; // bottom-left
	num += mCellArray[y*caSize + xIdxNeg].state; // middle-left

	return num;

}