#include <windows.h>
#include <gl/glut.h>
#include <iostream>
#include <vector>

#include "brushman.h"
#include "vector3.h"

/*
#include <il/il.h>
#include <il/ilu.h>
#include <il/ilut.h>
*/
#include "particle.h"

#include "timer.h"

#include <math.h>

using namespace std;

vector<int> img;

GLubyte* imgData;
float* paperField;

CPBrushManager brushMan;

Vec3 cGravity(0, -0.3, -0.1);
float cEvap = 0.001;
float cFric = .5;
int drawpaper = 0;

CTimer timer;
float dt=0;

int leftButton = 0;
int rightButton = 0;

int mousex, mousey;
int lmousex, lmousey;

colCMY selCol(0.4, 0.4, 0.05);


GradFieldCell	gradField[600*600];
float			wetnessArray[600*600];

float colArr[600*600*3];

TEXTUREIMAGE paptex;


PPSystem ppSys(cEvap, cGravity, cFric);


void clearWetness()
{
	for (int i = 0; i < 360000; i++)
	{
		wetnessArray[i] = 0;
	}
}

void clearColour()
{
	for (int i = 0; i < 600*600*3; i+=3)
	{
		colArr[i] = 1.0f;
		colArr[i+1] = 1.0f;
		colArr[i+2] = 1.0f;
	}
}

void dry(float dt)
{
	for (int i = 0; i < 360000; i++)
	{
		if (wetnessArray[i] == 0) continue;
		wetnessArray[i] -= cEvap*dt;
		if (wetnessArray[i] < 0) wetnessArray[i] = 0;
	} 
}

int first = 1;

void radialSpray(int x, int y, colCMY col)
{
	for (int i = 0; i < 2; i++)    
	{
		float t = 1.0f / (float) i;
		float a = t * TWOPI;
		float cosa = cos(a);
		float sina = sin(a);
		Vec3 v(-sina, cosa, 0);
		v*= 5;
		Vec3 p(x, y, 1);
		ppSys.addParticle(1, 1, p, v, col);
	}
	for (int i = 0; i < 2; i++)    
	{
		float t = 1.0f / (float) i;
		float a = t * TWOPI;
		float cosa = cos(a);
		float sina = sin(a);
		Vec3 v(sina, -cosa, 0);
		v*= 1;
		Vec3 p(x, y, 1);
		ppSys.addParticle(1, 1, p, v, col);
	}
}

void stamp(int x, int y, int size, colCMY col)
{
	int o = 1;
	int o2 = 1;
	int space = 10;
	for (int j = -size; j < size; j+=space)
	{
		o = !o;
		for (int i = -size; i < size; i+=space)
		{
			if (i >= -10 && i <= 10 && j >= -10 && j <= 10) continue;

			if (o)
			{
				if (o2)
				{
					Vec3 v((x+i)-x, (y+j)-y, 0);
					v*=0.2;
					//v*=100;
					Vec3 p(x+i+space/2, y+j+space/2, 0);
					ppSys.addParticle(1, 1, p, v, col);
					o2 = 0;
				}
				else
				{
					Vec3 v((x+i)-x, (y+j)-y, 0);
					v*=0.2;
					//v*=100;
					Vec3 p(x+i+space/2, y+j, 0);
					ppSys.addParticle(1, 1, p, v, col);
					o2 = 1;
				}
				
			}
			else
			{
				if (o2)
				{
					Vec3 v((x+i)-x, (y+j)-y, 0);
					v*=0.2;
					//v*=100;
					Vec3 p(x+i+space/2, y+j+space/2, 0);
					ppSys.addParticle(1, 1, p, v, col);
					o2 = 0;
				}
				else
				{
					Vec3 v((x+i)-x, (y+j)-y, 0);
					v*=0.2;
					//v*=100;
					Vec3 p(x+i+space/2, y+j, 0);
					ppSys.addParticle(1, 1, p, v, col);
					o2 = 1;
				}
				
			}
		}
	}

	for (int j = -size; j < size; j++)
	{
		for (int i = -size; i < size; i++)
		{
			wetnessArray[(j+y)*600+(i+x)] += 1;
		}
	}
}

void flat(int mx, int my, int lmx, int lmy, colCMY c, float s)
{
	//work out the direction of the brush stroke
	Vec3 dir;
	Vec3 pos(mx, my, 0);
	dir.x = mx - lmx;
	dir.y = my - lmy;

	// get the normal 
	Vec3 norm = dir ^ Vec3(0, 0, 1);
	Vec3 top = norm.normal() * s/2;
	Vec3 bottom = -norm.normal() * s/2;

	Vec3 pt = pos + top;
	Vec3 pb = pos + bottom;

	int steps = (rand() % 10) + 10;
	for (int t = 0; t < steps; t++)
	{
		Vec3 pnt = pt * (1-t/(float)steps) + pb * t/(float)steps;
		Vec3 v = pnt - pos;
		v*=3;
		ppSys.addParticle(1, 1, pnt, v, c);
	}
}

void display()
{
	timer.stoptimer();
	dt = fabs(timer.timeintv());
	timer.starttimer();
	
	if (first)
	{
		lmousex = mousex;
		lmousey = mousey;
		first = 0;
	}

	//dt*= 5;

	//if (first) 
	//{	
	//glClearColor(1, 0, 0, 0);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	//	first = 0;
	//}
	//glBindTexture(GL_TEXTURE_2D, brushMan.brushArray[0].glID);
	//glRasterPos2i(0,0);
	
	// if the left button is down add some particles
	if (leftButton)
	{
		float md = sqrt((mousex-lmousex)*(mousex-lmousex) + (mousey-lmousey)*(mousey-lmousey));
		if (md >= 1) 
		{
			for (int i = 1; i <= (int)md; i++)
			{
				float t = (float)i / md;
				float mx = (1-t)*lmousex + t*mousex;
				float my = (1-t)*lmousey + t*mousey;
				
					//Vec3 p(mousex, mousey, 1);
					//Vec3 v(0, 10, 0);
					colCMY c(0.4,0.4,0.05);
					//ppSys.addParticle(1, 1, p, v, c);
					/*
					for (int j = -30; j < 30; j+=5)
					{
						for (int i = -30; i < 30; i+=5)
						{
							radialSpray((int)mx+i, (int)my+j, c);
						}
					}*/
					stamp(mx, my, 30, selCol);
					//flat(mx, my, lmousex, lmousey, c, 30);

					//radialSpray(mousex, mousey, c);
				
			//cout << ppSys.numParticles() << "\n";
			}
		}
		lmousex= mousex;
		lmousey = mousey;
	}
	else
	{
		lmousex= mousex;
		lmousey = mousey;

	}
	
	ppSys.draw(colArr, dt, wetnessArray);
	ppSys.update(dt, gradField, wetnessArray, colArr);
	
	
	

	glDrawPixels(600, 600, GL_RGB, GL_FLOAT, colArr);

	//ppSys.drawP();
	

	if (drawpaper) glDrawPixels(600, 600, GL_RGB, GL_FLOAT, paperField);
	// dry the paper out 
	dry(dt);
	
	glutSwapBuffers();
}



void initGl()
{
	
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	//glViewport(0, 0, 600, 600);
	gluOrtho2D(0, 600, 0, 600);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glClearColor(1, 1, 1, 1);
}

float getHeight(int x, int y)
{
	if (x < 0 || x >= 600 || y < 0 || y >= 600)
	{
		cout << "Bad index: " << x << " " << y << endl;
		return -1;		
	}

    return paperField[y*600 + x];
}

void init()
{
	/*
	int tmp = ilutGLLoadImage("./pap_hotPress.tga");
	img.push_back(tmp);

	int iw = ilGetInteger(IL_IMAGE_WIDTH);
	int ih = ilGetInteger(IL_IMAGE_HEIGHT);

	cout << "iw: " << iw << " ih " << ih << endl;

	iw = 154;
	ih = 154;

	imgData = new ILubyte[iw*ih*3];
	//ilCopyPixels(0, 0, 0, iw, ih, 0, IL_RGB, IL_FLOAT, imgData);
	imgData = ilGetData();

	for (int i = 0; i < iw*ih*3; i+=3)
	{
		cout << (int)imgData[i] << " " << (int)imgData[i+1] << " " << (int)imgData[i+2] << endl;
	}*/
	char filename[256] = "./pap_hotPress.tga";
	if (loadTGA(&paptex, filename))
	{
		cout << "Loaded paper " << filename << endl;
		cout << "width: "	<< paptex.width << endl;
		cout << "height: "	<< paptex.height << endl;
	}

	int iw = paptex.width;
	int ih = paptex.height;

	imgData = paptex.imageData;

	

	paperField = new float[600*600*3];
	
	int imx = 0, imy = 0;
	for (int y = 0; y < 600; y++, imy++)
	{
		if (imy >= ih) imy = 0;
		for (int x = 0; x < 600; x++, imx++)
		{
			if (imx >= iw) imx = 0;
			

			paperField[(y*600 + x)*3] = (float)imgData[(imy*iw + imx)*3] / 255.0f;
			paperField[(y*600 + x)*3+1] = (float)imgData[(imy*iw + imx)*3+1] / 255.0f;
			paperField[(y*600 + x)*3+2] = (float)imgData[(imy*iw + imx)*3+2] / 255.0f;
		}	
	}

	// now init the gradient field
	for (int y = 0; y < 600; y++)
	{
		for (int x = 0; x < 600; x++)
		{
			float thisHeight = getHeight(x, y);
			float bottomheight, topheight, leftheight, rightheight;
			if (x == 0)
				leftheight = thisHeight;
			else
				leftheight = getHeight(x-1, y);
			if (x == 599)
				rightheight = thisHeight;
			else
				rightheight = getHeight(x+1, y);
			if (y == 0)
				bottomheight = thisHeight;
			else
				bottomheight = getHeight(x, y-1);
			if (y == 599)
				topheight = thisHeight;
			else
				topheight = getHeight(x, y+1);

			Vec3	topPos(0, 1, topheight), 
					rightPos(1, 0, rightheight), 
					leftPos(-1, 0, leftheight), 
					bottomPos(0, -1, bottomheight), 
					thisPos(0, 0, thisHeight);

			gradField[y*600+x].top = topPos - thisPos;
			gradField[y*600+x].bottom = bottomPos - thisPos;
			gradField[y*600+x].left = leftPos - thisPos;
			gradField[y*600+x].right = rightPos - thisPos;

			gradField[y*600+x].av = (	gradField[y*600+x].top*fabs(gradField[y*600+x].top.z) + 
										gradField[y*600+x].bottom*fabs(gradField[y*600+x].bottom.z)  + 
										gradField[y*600+x].left*fabs(gradField[y*600+x].left.z)  + 
										gradField[y*600+x].right*fabs(gradField[y*600+x].right.z) ) / 4;

			//cout << gradField[y*600+x].av << endl;

			// gradfield should now contain the gradient field for the paper
			
			

		}
	}

}

void keyup(unsigned char key, int x, int y)
{
	switch (key)
	{
	case 'p':
		drawpaper = !drawpaper;
		break;
	case 'q':
		selCol = colCMY(0, 0, 0);
		break;
	case 'w':
		selCol = colCMY(.4, .4, 0.05);
		break;
	case 'e':
		selCol = colCMY(.05, .4, .4);
		break;
	case 'r':
		selCol = colCMY(.05, .2, .6);
		break;
	}
}

void mouse(int button, int state, int x, int y)
{
	switch (button)
	{
	case GLUT_LEFT_BUTTON:
		if (state == GLUT_DOWN) leftButton = 1;
		if (state == GLUT_UP) leftButton = 0;
		break;
	default:
		break;
		
	}
	
	//lmousex = mousex;
	//lmousey = mousey;

	mousex = x;
	mousey = 600-y;
}

void mousem(int x, int y)
{
	//lmousex = mousex;
	//lmousey = mousey;
	mousex = x;
	mousey = 600-y;
}

void pmouse(int x, int y)
{
	//lmousex = mousex;
	//lmousey = mousey;
	mousex = x;
	mousey = 600-y;
}

void main (int argc, char** argv)
{
	glutInit(&argc, argv);
	glutInitWindowSize(600, 600);
	glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);
	glutInitWindowPosition(100, 100);
	glutCreateWindow("ec");
	glutDisplayFunc(display);
	glutIdleFunc(display);
	glutKeyboardUpFunc(keyup);
	glutMouseFunc(mouse);
	glutMotionFunc(mousem);
	glutPassiveMotionFunc(pmouse);

	glEnable(GL_TEXTURE_2D);
	//glEnable(GL_BLEND);
	//glDisable(GL_DEPTH_TEST);
	//glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	//glShadeModel(GL_SMOOTH);

	initGl();
	
	
/*
	ilInit();
	iluInit();
	ilutRenderer(ILUT_OPENGL);*/

	init();

	//brushMan.ilLoadBrushes();
	
	clearColour();

	timer.starttimer();



	glutMainLoop();
}