#include "particle.h"

void PP::drawParticle(float size)
{
	float s = size * 0.5;
	glColor3f(1-colour.c, 1-colour.m, 1-colour.y );
	glVertex2f(position.x-s, position.y-s);
	glVertex2f(position.x+s, position.y-s);
	glVertex2f(position.x+s, position.y+s);
	glVertex2f(position.x-s, position.y+s);
}






unsigned PPSystem::numParticles()
{
	return (unsigned)particleArray.size();
}


void PPSystem::kill(int idx)
{
	particleArray[idx].alive = 0;
	particleArray[idx].killedThisStep = 1;
	deadList.push_front(idx);
}


void PPSystem::reincarnate(float _mass, float _wetness, Vec3 _position, Vec3 _velocity, colCMY _colour)
{
	int idx = *deadList.begin();
	deadList.pop_front();

	particleArray[idx].alive = 1;
	particleArray[idx].mass = _mass;
	particleArray[idx].killedThisStep = 0;
	particleArray[idx].wetness = _wetness;
	particleArray[idx].position = _position;
	particleArray[idx].velocity = _velocity;
	particleArray[idx].colour = _colour;

}


void PPSystem::addParticle(float _mass, float _wetness, Vec3 _position, Vec3 _velocity, colCMY _colour)
{
	//cout << deadList.size() << "\n";
	size_t s = deadList.size();
	// if no particles to reincarnate
	if (s == 0)
	{
		PP pp(_mass, _wetness, _position, _velocity,_colour);
		particleArray.push_back(pp);
	}
	else
		reincarnate(_mass, _wetness,_position,_velocity,_colour);
}


void PPSystem::drawParticles(float size)
{
	vector<PP>::iterator it = particleArray.begin();
	for (; it != particleArray.end(); ++it)
		it->drawParticle(size);
}

inline
void colourPixel(int x, int y, colCMY col, float* colArray)
{
	if (x < 0 || x >= UBOUND || y < 0 || y >= UBOUND) return;

	colArray[(y*600+x)*3]	-= col.c;
	if (colArray[(y*600+x)*3] < 0) colArray[(y*600+x)*3] = 0;
	if (colArray[(y*600+x)*3] > 1) colArray[(y*600+x)*3] = 1;
	colArray[(y*600+x)*3+1] -= col.m;
	if (colArray[(y*600+x)*3+1] < 0) colArray[(y*600+x)*3+1] = 0;
	if (colArray[(y*600+x)*3+1] > 1) colArray[(y*600+x)*3+1] = 1;
	colArray[(y*600+x)*3+2] -= col.y;
	if (colArray[(y*600+x)*3+2] < 0) colArray[(y*600+x)*3+2] = 0;
	if (colArray[(y*600+x)*3+2] > 1) colArray[(y*600+x)*3+2] = 1;
}

void PPSystem::draw(float* colArray, float dt, float* w)
{
	
	//glColor3f(0, 0, 0);
	//glBegin(GL_POINTS);
	vector<PP>::iterator p = particleArray.begin();
	for (; p != particleArray.end(); ++p)
	{
		if (p->alive ) 
		{
			int x = (int)p->position.x;
			int y = (int)p->position.y;
			//glColor3f(1-p->colour.c, 1-p->colour.m,1-p->colour.y);
			//glVertex2f(p->position.x, p->position.y);	
			// absorb some colour into paper
			if ((p->colour.c > 0 || p->colour.m > 0 || p->colour.y > 0) && w[y*600+x] > 0)
			{
				colCMY colt;
				colt = p->colour * 1 *dt;
				p->colour -= colt;
				colourPixel(x,y, colt, colArray);
			}

			//pickup some colour from the paper
			colCMY colp;
			colp.c = colArray[(y*600+x)*3];
			colp.m = colArray[(y*600+x)*3+1];
			colp.y = colArray[(y*600+x)*3+2];
			if (colp.c > 0 && colp.m > 0 && colp.y > 0)
			{
				colp *= 0.1 *dt;
				p->colour += colp;
				if (colp.c > 1)colp.c = 1;
				if (colp.m > 1)colp.m = 1;
				if (colp.y > 1)colp.y = 1;
				colp.c = -colp.c;
				colp.m = -colp.m;
				colp.y = -colp.y;
				colourPixel(x, y, colp,colArray);
			}


		}
	}
	//glEnd();
}

void PPSystem::drawP()
{

	glColor3f(0, 0, 0);
	glBegin(GL_POINTS);
	vector<PP>::iterator p = particleArray.begin();
	for (; p != particleArray.end(); ++p)
	{
		if (p->alive && (p->colour.c > 0 || p->colour.m > 0 || p->colour.y > 0)) 
		{
			//int x = (int)p->position.x;
			//int y = (int)p->position.y;
			glColor3f(1-p->colour.c, 1-p->colour.m,1-p->colour.y);
			glVertex2f(p->position.x, p->position.y);	
			//colCMY colt;
			//colt = p->colour * 0.5 *dt;
			//p->colour -= colt;
			//colourPixel(x,y, colt, colArray);
		}
	}
	glEnd();
}


void PPSystem::update(float dt, GradFieldCell* gf, float* w, float* colArr)
{
	
	for (int i = 0; i < particleArray.size(); ++i)
	{
		// if this particle's dead, ignore it
		if (!particleArray[i].alive) continue;

		// check for kill conditions
		if (particleArray[i].position.x < 0 ||
			particleArray[i].position.y < 0 ||
			particleArray[i].position.x >= UBOUND ||
			particleArray[i].position.y >= UBOUND) 
		{
			kill(i);
			continue;
		}
		
		int x = (int)particleArray[i].position.x;
		int y = (int)particleArray[i].position.y;
		
		// apply paper gradient forces
		particleArray[i].velocity += (gf[y*600+x].av*1000);

		// add friction
		particleArray[i].velocity *= 1-cFric*dt;
		// if the wetness of the paper is 0, setthe velocity to 0;
		//if (w[y*600+x] == 0)
		//{
		//	Vec3 tmp(0, 0, 0);
		//	particleArray[i].velocity = tmp;
		//}
		particleArray[i].velocity*= w[y*600+x];

		//gravity
		particleArray[i].velocity+= cGrav;

		//scale by wetness
		particleArray[i].velocity *= particleArray[i].wetness;

		
		// 
		particleArray[i].velocity.truncate(100);

		// update its position and velocity
		particleArray[i].lastFramePos = particleArray[i].position;
		particleArray[i].position += particleArray[i].velocity *dt;
		// constrain the partile to the paper
		if (particleArray[i].position.z > 1) particleArray[i].position.z = 1;
		else 
			if (particleArray[i].position.z < 0) particleArray[i].position.z = 0;

		
		
		
		// dry the particle out
		//particleArray[i].wetness -= cEvap*dt;
		//if (particleArray[i].wetness < 0) particleArray[i].wetness = 0;
		// absorb some wetness into the paper
		particleArray[i].wetness -= cEvap*dt;
		if (particleArray[i].wetness < 0) particleArray[i].wetness = 0;
		w[y*600+x] += cEvap*dt;

		// check for kill conditions
		if (particleArray[i].position.x < 0 ||
			particleArray[i].position.y < 0 ||
			particleArray[i].position.x >= UBOUND ||
			particleArray[i].position.y >= UBOUND) kill(i);

		if (particleArray[i].velocity.length() < 0.01) kill(i);
		if (particleArray[i].wetness == 0) kill(i);

	}

}